// SilberLand MUDlib
//
// p/daemon/ifingerd.c -- Internet finger request daemon (x-erq required)
//
// $Date: 1997/11/09 19:18:17 $
/* $Revision: 1.1 $
 * $Log: ifingerd.c,v $
 * Revision 1.1  1997/11/09 19:18:17  Woody
 * Initial revision
 *
 */

#include "socket.h"
#include <wizlevels.h>
#include <errno.h>

#define LUSER      0
#define RUSER      1
#define HOST       2
#define CONTINUED  3

#define IF_PORT    79   // the standard unix finger port
#define IF_TIMEOUT 90   // timeout in seconds
#define IF_MAXSOCK 10   // maximum number of pending requests

mapping sockets;

void callback(int fd, int act, mixed a, mixed b);

int create()
{
  sockets = allocate_mapping(0, 4);
}

int ifinger(string str)
{
  string user, host;
  int sock;

  if (!objectp(this_player())) return 0;
  if (this_player()!=this_interactive()) return 0;
  if (!IS_LORD(this_player()))
    return notify_fail("You are not allowed to use ifinger.\n"),0;
  if (sizeof(sockets) > IF_MAXSOCK)
    return notify_fail("Too many pending ifinger requests, retry later.\n"),0;
  if (!str) str="";
  host = __HOST_NAME__;
  user = str;
  sscanf(str, "%s@%s", user, host);

  sock = SOCKETD->socket_connect(host, 79, #'callback);
  sockets[sock, LUSER] = this_player();
  sockets[sock, RUSER] = user;
  sockets[sock, HOST]  = host;
  // printf("[finger to: %s@%s]\n", user, host);
  call_out("time_out", IF_TIMEOUT, sock);
  return 1;
}

void callback(int sock, int act, mixed a, mixed b)
{
  if (!sockets[sock, LUSER] && act != SOCKET_CLOSE) {
    SOCKETD->socket_close(sock);
    return;
  }
  switch(act) {
    case SOCKET_READY:
      SOCKETD->socket_write(sock, sprintf("%s\r\n", sockets[sock, RUSER]));
      break;
    case SOCKET_READ:
      if (!sockets[sock, CONTINUED])
      {
        tell_object(sockets[sock, LUSER], sprintf("[%s@%s]\n",
          sockets[sock, RUSER], sockets[sock, HOST]));
        sockets[sock, CONTINUED] = 1;
      }
      tell_object(sockets[sock, LUSER], a);
      break;
    case SOCKET_CLOSE:
      tell_object(sockets[sock, LUSER], "\n");
      efun::m_delete(sockets, sock);
      break;
    case SOCKET_ERROR: {
      string err;
      switch(b) {
        case ECONNREFUSED:
          err = "Connection refused";
          break;
        case EHOSTUNREACH:
          err = "Host unreachable";
          break;
        case ENETUNREACH:
          err = "Net unreachable";
          break;
        default:
          err = sprintf("Code %d (erq errorcode %d)", b, a);
          break;
      }
      tell_object(sockets[sock, LUSER],
        sprintf("ifinger: Error in connection to %s: %s.\n",
          sockets[sock, HOST], err));
      efun::m_delete(sockets, sock);
    }
    break;
  }
  return;
}

void help()
{
  write("Usage: ifinger [[<user>]@<host>]\n"
        "Send an internet finger request.\n");
  return;
}

void time_out(int sock)
{
  if (member(sockets,sock))
  {
    if (objectp(sockets[sock, LUSER]))
      tell_object(sockets[sock, LUSER],
        sprintf("ifinger to %s@%s timed out.\n",
          sockets[sock, RUSER], sockets[sock, HOST]));
    efun::m_delete(sockets, sock);
    return;
  }
}

mapping query_sockets() {
  return copy_mapping(sockets);
}

